"use strict";
this.name        = "miningcontracts";
this.author      = "Diagoras";
this.copyright   = "2015 Maxim Savchenko";
this.licence     = "Ms-PL";
this.description = "Excavate asteroids for reward with deadline conditions.";

this.startUp = function() {
	this.MC_contracts = [];

	if (missionVariables.MC_initialized)
	{
		this.MC_system = missionVariables.MC_system;
		this.MC_clock = missionVariables.MC_clock;
		this.MC_debt = missionVariables.MC_debt;
		var s = missionVariables.MC_contracts;
		if (s && s.length > 0) this.MC_contracts = JSON.parse(s);
	}
	else
	{
		this.MC_system = "";
		this.MC_clock = 0;
		this.MC_debt = 0;
	}
};

this.startUpComplete = function() {
	this.MC_countAsteroids();
	this.MC_setInstructions();
	if(player.ship.docked) this.shipDockedWithStation(player.ship.dockedStation);
};

this.shipDockedWithStation = function(station)
{
	if (station.shipClassName != "Rock Hermit") return;

	if (this.MC_system != system.name)
	{
		this.MC_system = system.name;
		this.MC_clock = 0;
		this.MC_contracts = [];
		this.MC_debt = 0;
	}

	this.MC_generateContracts();

	station.setInterface("MiningContracts-bm", {
		title: "Mining Contracts",
		category: "Contracts",
		summary: "Sign up for a local mining operations.",
		callback: this.MC_Interface.bind(this)
	});
};

this.playerWillSaveGame = function(message)
{
	missionVariables.MC_initialized = 1;
	missionVariables.MC_system = this.MC_system;
	missionVariables.MC_clock = this.MC_clock;
	missionVariables.MC_debt = this.MC_debt;
	missionVariables.MC_contracts = JSON.stringify(this.MC_contracts);
};

this.shipExitedWitchspace = function()
{
	this.MC_countAsteroids();

	var number = this.MC_contracts.length;
	var total = this.MC_debt;

	for (var i = 0; i < number; i++)
	{
		var contract = this.MC_contracts[i];
		if (contract.signed) total += contract.amount * contract.penalty - contract.reward;
	}

	this.MC_contracts = [];
	this.MC_debt = 0;
	this.MC_clock = 0;

	if (total > 0)
	{
		player.bounty += 50;
		player.commsMessage(
			"Mining Association of " +
			this.MC_system +
			" have offered bounty of " +
			formatCredits(50, true, true) +
			" for your head");
	}

	this.MC_setInstructions();
};

this.MC_countAsteroids = function()
{
	this.MC_icequota = system.countShipsWithRole("staer9_asteroid") / system.countShipsWithRole("asteroid");
}

this.MC_setInstructions = function()
{
	var total = {
		alloys: 0,
		minerals: 0
	};

	var deadline = {
		alloys: Infinity,
		minerals: Infinity
	};

	var number = this.MC_contracts.length;

	for (var i = 0; i < number; i++)
	{
		var contract = this.MC_contracts[i];
		if (contract.signed == 1)
		{
			total[contract.cargo] += contract.amount;
			deadline[contract.cargo] = Math.min(deadline[contract.cargo], contract.deadline);
		}
	}

	var flags = {
		alloys: total.alloys > 0,
		minerals: total.minerals > 0
	};

	function msg(cargo)
	{
		return flags[cargo] ? total[cargo] + "t of " + cargo + " (" + MC_clockStringToMinutes(deadline[cargo]) + ")" : "";
	};

	if (flags.alloys || flags.minerals)
	{
		this.MC_active = true;
		var andMsg = flags.alloys && flags.minerals ? " and " : "";
		mission.setInstructions("Mining Contracts: " + msg("alloys") + andMsg + msg("minerals"));
	}
	else
	{
		this.MC_active = false;
		mission.setInstructions(null);
	}
};

function MC_clockStringToMinutes(time)
{
	return clock.clockStringForTime(time).slice(4,13);
}

this.MC_Interface = function()
{
	var debtMsg =
		this.MC_debt > 0 ? "\n\nYou are in debt to Mining Association of " + this.MC_system +
		" (" + formatCredits(this.MC_debt, true, true) + ")" : "";

	var number = this.MC_contracts.length;

	var parameters = {
		title: "Mining Contracts",
		message:
			"Rock Hermit main lounge is as dusty as everything in this place. " +
			"Couple grumpy miners are playing cards at the only table in the corner. " +
			"Balding clerk is sleeping at the front desk of local Mining Association. " +
			"Large blackboard with listing of current contracts is hanging over him. " +
			"Suddenly ancient fax machine standing on the desk squeals. " +
			"Dragged out of the hay clerk swears, wipes one line from a blackboard and falls back to sleep. " +
			"On the opposite wall you notice numerous square hatches layered in two rows under flaked advert:\n\n" +
			"RESTING PODS! NO TIME LIMIT! ONLY " + formatCredits(3, true, true) + " PER USE!" +
			debtMsg,
		choices: {
			"WAIT": "Wait for new contract (" + formatCredits(3, true, true) + ")",
			"XXIT": "Exit Mining Contracts"
		},
		allowInterrupt: true
	};

	for (var i = 0; i < number; i++)
	{
		var contract = this.MC_contracts[i];
		var status = ["OFFER", "TAKEN", "BLOWN"][contract.signed];
		parameters.choices["C" + i]
			= status + ": "
			+ contract.amount + "t of "
			+ contract.cargo + " before "
			+ MC_clockStringToMinutes(contract.deadline) + " for "
			+ formatCredits(contract.reward, true, true);
	}

	if (this.MC_debt > 0) parameters.choices["DEBT"] = "Repay debt";

	mission.runScreen(parameters, this.MC_contractDetails.bind(this));
};

this.MC_contractDetails = function(choice)
{
	if (choice == null) return;

	if (choice == "DEBT")
	{
		if (this.MC_debt > player.credits)
		{
			player.consoleMessage("NOT ENOUGH CREDITS");
			this.MC_Interface();
		}
		else
		{
			player.commsMessage(formatCredits(this.MC_debt, true, true) + " deducted");
			player.credits -= this.MC_debt;
			this.MC_debt = 0;
			this.MC_Interface();
		}
		return;
	}

	if (choice == "WAIT")
	{
		if (player.credits < 3)
		{
			player.consoleMessage("NOT ENOUGH CREDITS");
			this.MC_Interface();
			return;
		}

		player.commsMessage(formatCredits(3, true, true) + " deducted");
		player.credits -= 3;
		this.MC_waitForContract();
		this.MC_Interface();
		return;
	}

	if (choice == "XXIT") return;

	var cid = choice.replace("C", "").split(":");
	var contract = this.MC_contracts[cid[0]];
	var debt = this.MC_debt;
	var header, descr, deliv, ps, id, option;
	var mediator = "Mining Association of " + this.MC_system;
	var warning = "\n\nn.b. " + mediator + " reserves the right to set a bounty on heads of defaulting contractors.";

	var terms =
		"Terms and conditions:\n" +
		"Payment is calculated and carried out after one of three events - " +
		"full delivery of required cargo before due time, " +
		"premature contract termination initiated by contractor or " +
		"coming of due time. " +
		"If forefeit penalty exceeds fee, the difference becomes contractor's debt liability to " + mediator + ". " +
		"Contract fee meant for indebted contractor will be detained for repayment first. " +
		"Contractor leaving system before completion of obligations or debt retirement will be recognised as defaulter." +
		warning;

	function debtMsg(reward)
	{
		return debt > 0 ?
			"\n\nDue to unpaid debt to " + mediator + " we are withholding " +
			formatCredits(Math.min(reward, debt), true, true) +
			" from your reward." : "";
	}

	switch (contract.signed)
	{
		case 0:
			header = "PUBLIC CONTRACTUAL OFFER FOR MINING OPERATIONS";
			descr = "Subject of delivery";
			deliv = contract.amount;
			ps = terms;
			id = "SIGN";
			option = "Sign contract";
			break;

		case 1:
			descr = "Delivered";
			deliv = (contract.gross - contract.amount) + "/" + contract.gross + " ";

			if (contract.amount > 0)
			{
				header = "OPERATING CONTRACT";
				ps = terms;
				id = "CARGO";
				option = "Dispatch cargo (" + Math.min(contract.amount, manifest[contract.cargo]) + "t)";
			}
			else
			{
				header = "CONTRACT COMPLETION";
				ps = 
					"You have successfully performed a contract. " +
					"According to the terms your reward is " + formatCredits(contract.reward, true, true) + ". " +
					"We are looking forward to working with you in the future." + debtMsg(contract.reward);
				id = "REWARD";
				option = "Collect reward";
			}

			break;

		case 2:
			header = "TERMINATED CONTRACT";
			descr = "Delivered";
			deliv = (contract.gross - contract.amount) + "/" + contract.gross + " ";
			var balance = contract.reward - contract.amount * contract.penalty;

			if (balance < 0)
			{
				ps = 
					"You have failed to perform a contract and forefeit penalty exceeds your fee. " +
					"According to the terms your debt to " + mediator + " is increased by " + formatCredits(-balance, true, true) + ". " +
					"This debt may be liquidated with cash deposit. " +
					"Until full debt liquidation all your future contract fees will be detained for repayment first. " +
					"If you leave system before full debt liquidation, you will be recognised as defaulter. " +
					"Please, do not delay your payments." +
					warning;
				id = "DEBT";
				option = "Take debt notification";
			}
			else
			{
				ps = 
					"You have failed to perform a contract fully and forefeit penalty have been deducted from your fee. " +
					"According to the terms your reward is " + formatCredits(balance, true, true) + ". " +
					"Please, be more diligent next time." + debtMsg(balance);
				id = "REWARD";
				option = "Collect reward";
			}

			break;
	}

	var parameters = {
		title: "Mining Contracts",
		message:
			header + "\n" +
			"\n" +
			"Ordering party: " + contract.company + "\n" +
			"Mediator: Mining Association of " + this.MC_system + "\n" +
			descr + ": " + deliv + "t of " + contract.cargo + "\n" +
			"Due time: " + clock.clockStringForTime(contract.deadline) + "\n" +
			"Fee: " + formatCredits(contract.reward, true, true) + "\n" +
			"Forfeit penalty: " + formatCredits(contract.penalty, true, true) + " per ton of shortage\n" +
			"\n" +
			ps,
		choices: {},
		allowInterrupt: id != "REWARD" && id != "DEBT"
	};

	parameters.choices["0_" + id + ":" + cid[0]] = option;
	var next = "2_NEXT:" + cid[0];
	var prev = "3_PREV:" + cid[0];

	if (parameters.allowInterrupt)
	{
		if (id == "CARGO") parameters.choices["1_TERMINATE:" + cid[0]] = "Terminate contract";

		if (contract.signed == 0)
		{
			parameters.choices[next] = "Next unsigned contract";
			parameters.choices[prev] = "Previous unsigned contract";
		}
		else if (contract.signed == 1)
		{
			parameters.choices[next] = "Next unfinished contract";
			parameters.choices[prev] = "Previous unfinished contract";
		}

		parameters.choices["4_BACK"] = "Back to the list";
		if (id == "SIGN") parameters.initialChoicesKey = "4_BACK";
	}

	switch (cid[1])
	{
		case "N":
			parameters.initialChoicesKey = next;
			break;

		case "P":
			parameters.initialChoicesKey = prev;
			break;
	}

	mission.runScreen(parameters, this.MC_contractAction.bind(this));
};

this.MC_contractAction = function(choice)
{
	if (choice == null) return;
	var ch = choice.split(":");

	switch (ch[0])
	{
		case "0_SIGN":
			var idx = ch[1];
			this.MC_contracts[idx].signed = 1;
			this.MC_contracts[idx].gross = this.MC_contracts[idx].amount;
			player.consoleMessage("CONTRACT SIGNED");
			this.MC_setInstructions();
			this.MC_contractDetails("C" + idx);
			break;

		case "0_DEBT":
			var idx = ch[1];
			var contract = this.MC_contracts[idx];
			var debt = contract.amount * contract.penalty - contract.reward;
			this.MC_debt += debt;
			this.MC_contracts.splice(idx, 1);
			this.MC_Interface();
			break;

		case "0_REWARD":
			var idx = ch[1];
			var contract = this.MC_contracts[idx];
			var reward = contract.reward - contract.amount * contract.penalty;
			var repayment = Math.min(reward, this.MC_debt);
			reward -= repayment;
			this.MC_debt -= repayment;
			player.commsMessage(formatCredits(reward, true, true) + " received");
			player.credits += reward;
			this.MC_contracts.splice(idx, 1);
			this.MC_Interface();
			break;

		case "0_CARGO":
			var idx = ch[1];
			var contract = this.MC_contracts[idx];
			var amount = Math.min(contract.amount, manifest[contract.cargo]);

			if (amount > 0)
			{
				player.commsMessage(amount + "t of " + contract.cargo + " dispatched");
				this.MC_contracts[idx].amount -= amount;
				manifest[contract.cargo] -= amount;
				this.MC_setInstructions();
			}
			else player.consoleMessage("no " + contract.cargo + " on board");

			this.MC_contractDetails("C" + idx);
			break;

		case "1_TERMINATE":
			var idx = ch[1];
			var contract = this.MC_contracts[idx];

			var parameters = {
				title: "Mining Contracts",
				message: "Are you sure you want to prematurely terminate contract with " + contract.company + "?",
				choices: {},
				allowInterrupt: true
			};

			parameters.choices["0_TERMYES:" + idx] = "Yes";
			parameters.choices["1_TERMNO:" + idx] = "No";
			parameters.initialChoicesKey = "1_TERMNO:" + idx;
			mission.runScreen(parameters, this.MC_contractAction.bind(this));
			break;

		case "0_TERMYES":
			var idx = ch[1];
			this.MC_contracts[idx].signed = 2;
			player.consoleMessage("CONTRACT TERMINATED");
			this.MC_setInstructions();
			this.MC_contractDetails("C" + idx);
			break;

		case "1_TERMNO":
			var idx = ch[1];
			this.MC_contractDetails("C" + idx);
			break;

		case "2_NEXT":
			var idx = ch[1];
			var signed = this.MC_contracts[idx].signed;

			while (true)
			{
				idx++;
				if (idx >= this.MC_contracts.length) idx = 0;
				if (this.MC_contracts[idx].signed == signed) break;
			}

			this.MC_contractDetails("C" + idx + ":N");
			break;

		case "3_PREV":
			var idx = ch[1];
			var signed = this.MC_contracts[idx].signed;

			while (true)
			{
				if (idx <= 0) idx = this.MC_contracts.length;
				idx--;
				if (this.MC_contracts[idx].signed == signed) break;
			}

			this.MC_contractDetails("C" + idx + ":P");
			break;

		case "4_BACK":
			this.MC_Interface();
			break;
	}
};

function MC_randomElement(array)
{
	return array[Math.floor(Math.random() * array.length)];
}

function MC_randomExponential()
{
	return -Math.log(1 - Math.random());
};

function MC_randomNormals()
{
	var u = Math.sqrt(MC_randomExponential());
	var v = 2 * Math.PI * Math.random();

	return {
		x: u * Math.cos(v),
		y: u * Math.sin(v)
	};
};

function MC_ceilScale(value, scale)
{
	return scale * Math.ceil(value / scale);
};

function MC_floorScale(value, scale)
{
	return scale * Math.floor(value / scale);
};

this.MC_companyPatterns = [
	"? Company Limited",
	"? plc",
	"? S.A.",
	"? Group",
	"? Gold Corporation",
	"? Inc.",
	"Potash Corporation of ? Inc.",
	"? Copper Corporation",
	"MMC ? Nickel",
	"? Platinum Limited",
	"? Mining Corporation",
	"Aluminum Corporation of ? Limited",
	"? Platinum Holdings Limited",
	"? Mining Limited",
	"?Gold",
	"? Goldfields Corporation",
	"National Copper Corporation of ?",
	"? Industries Ltd."
];

this.MC_companyName = function()
{
	return MC_randomElement(this.MC_companyPatterns).replace("?", randomName());
};

this.MC_randomCargo = function()
{
	return Math.random() < this.MC_icequota ? "alloys" : "minerals";
};

function MC_logistic(x)
{
	return 1 / (1 + Math.exp(x));
};

function MC_uniform(x, y)
{
	return x + (y - x) * Math.random();
};

this.MC_amountMu = 3;
this.MC_timeMu = 6;
this.MC_generationRate = ((Math.exp(this.MC_amountMu) + 0.5) * Math.exp(this.MC_timeMu) + 450) / 4;

this.MC_createContract = function(time)
{
	var rn = MC_randomNormals();
	var amount = Math.ceil(Math.exp(rn.x + this.MC_amountMu));
	var deadline = MC_ceilScale(time + amount * Math.exp(rn.y / 2 + this.MC_timeMu), 900);
	var expires = MC_uniform(time, deadline);
	var cargo = this.MC_randomCargo();
	var price = player.ship.dockedStation.market[cargo].price / 10;
	var reward = MC_ceilScale(MC_floorScale(amount * price, 10) * MC_logistic(rn.y), 10);

	var contract = {
		company: this.MC_companyName(),
		expires: expires,
		deadline: deadline,
		cargo: cargo,
		amount: amount,
		reward: reward,
		penalty: price,
		signed: 0
	};

	this.MC_contracts.push(contract);
};

this.MC_generateContracts = function()
{
	var prevClock = this.MC_clock;
	var nextClock = clock.seconds;
	var pregen = 96 * 3600;
	if (nextClock - prevClock > pregen) prevClock = nextClock - pregen;

	while (true)
	{
		prevClock += MC_randomExponential() * this.MC_generationRate;
		if (prevClock > nextClock) break;
		this.MC_createContract(prevClock);
	};

	this.MC_clock = nextClock;
	this.MC_expireContracts();
};

this.MC_waitForContract = function()
{
	this.MC_clock += MC_randomExponential() * this.MC_generationRate;
	this.MC_createContract(this.MC_clock);
	var time = this.MC_clock - clock.seconds;
	if (time > 0) clock.addSeconds(time);
	this.MC_expireContracts();
};

this.MC_expireContracts = function()
{
	var number = this.MC_contracts.length;
	var updated = [];

	for (var i = 0; i < number; i++)
	{
		var contract = this.MC_contracts[i];
		if (contract.signed == 1 && contract.deadline < clock.seconds) contract.signed = 2;
		if (contract.signed || contract.expires > this.MC_clock) updated.push(contract);
	}

	this.MC_contracts = updated;
	this.MC_setInstructions();
};